import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver


def check_device_and_compute_regmean_2D(var, basins, e1t, e2t):
    """Function that checks device and calls computing functions.

    Checks if the computations are going performed in the CPU or the GPU:

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.

    Returns
    -------
    regmean: float32
        List containing regional mean for variable var
    """
    check_gpu = numba.cuda.is_available()
    if check_gpu is False:
        device = 'CPU'
        regmean = compute_regmean_2D_cpu(var, basins, e1t, e2t)
    else:
        use_gpu = True
        if use_gpu is True:
            device = 'GPU'
            regmean = compute_regmean_2D_gpu(var, basins, e1t, e2t)
        else:
            device = 'CPU'
            regmean = compute_regmean_2D_cpu(var, basins, e1t, e2t)
    print('Running on {}'.format(device))
    return regmean


def compute_regmean_2D_cpu(var, basins, e1t, e2t):
    """Function that computes the regional mean for 2D vars in the cpu.

       Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    times = var.shape[0]
    regmean_total = {}
    for basin, mask in basins.items():
        weights = compute_weights_2D(mask, e1t, e2t)
        regmean = np.empty(times)
        for t in range(times):
            regmean[t] = np.average(var[t, :, :], axis=(0, 1),
                                    weights=np.squeeze(weights))
        regmean_total[basin] = regmean
    return regmean_total


def compute_regmean_3D_levels(var, basins, e1t, e2t, e3t):
    """Function that computes the regional mean at every depth level
    for 3D vars in the cpu.

    Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.

    Returns
    -------
    regmean_total: float32
        List containing regional mean at every depth level for variable var.
    """
    times = var.shape[0]
    levs = var.shape[1]
    regmean_total = {}
    for basin, mask in basins.items():
        regmean = np.empty((times, levs))
        w = compute_weights_3D(mask, e1t, e2t, e3t)
        for t in range(times):
            for l in range(levs):
                regmean[t, l] = np.average(var[t, l, :, :], axis=(0, 1),
                                           weights=np.squeeze(w)[l, :, :])
        regmean_total[basin] = regmean
    return regmean_total


def compute_regmean_3D_cpu(var, basins, e1t, e2t, e3t):
    """Function that computes the regional mean for 3D vars in the cpu.

    Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.

    Returns
    -------
    regmean_total: float32
        List containing the regional mean for variable var.
    """
    times = var.shape[0]
    regmean_total = {}
    for basin, mask in basins.items():
        weights = compute_weights_3D(mask, e1t, e2t, e3t)
        regmean = np.empty(times)
        for t in range(times):
            regmean[t] = np.average(var[t, :, :, :], axis=(0, 1, 2),
                                    weights=np.squeeze(weights))
        regmean_total[basin] = regmean
    return regmean_total


def compute_regmean_2D_gpu(var, basins, e1t, e2t):
    """Function that computes the regional mean for 2D vars in the gpu.

    Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]
    return masks

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, lats*lons), dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_masks,
                                      gpu_weighted_var, gpu_vector_weight)

    regmean = {}
    for m in range(mask):
        reduced_weight = sum_red_cuda(gpu_vector_weight[m, :])
        regmean[m] = {}
        for t in range(times):
            reduced_var = sum_red_cuda(gpu_weighted_var[m, t, :])
            regmean[m][t] = reduced_var / reduced_weight

    return regmean


def compute_regmean_gpu_3D(var, basins, e1t, e2t, e3t):
    """Function that computes the regional mean for 3D vars in the gpu.

    Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.

    Returns
    -------
    regmean: float32
        List containing the regional mean for variable var.
    """
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)

    gpu_weighted_var = cuda.device_array((mask, times, levs*lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, levs*lats*lons),
                                          dtype=np.float32)

    variable_weight_cuda_3D[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_e3t,
                                         gpu_masks, gpu_weighted_var,
                                         gpu_vector_weight)

    regmean = {}
    for m in range(mask):
        reduced_weight = sum_red_cuda(gpu_vector_weight[m, :])
        regmean[m] = {}
        for t in range(times):
            reduced_var = sum_red_cuda(gpu_weighted_var[m, t, :])
            regmean[m][t] = reduced_var / reduced_weight

    return regmean


def compute_regmean_lev_gpu_3D(var, basins, e1t, e2t, e3t):
    """Function that computes the regional mean at every depth level
    for 3D vars in the cpu.

    Computes the weights for each region and performs a weighted average.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.

    Returns
    -------
    regmean: float32
        List containing regional mean at every depth level for variable var.
    """
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)

    gpu_weighted_var = cuda.device_array((mask, times, levs, lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, levs, lats*lons),
                                          dtype=np.float32)

    variable_weight_cuda_level_3D[grid, block](gpu_var, gpu_weights,
                                               gpu_weighted_var,
                                               gpu_vector_weight)
    del gpu_var, gpu_weights

    regmean = {}
    for m in range(mask):
        regmean[m] = {}
        for t in range(times):
            regmean[m][t] = {}
            for l in range(levs):
                reduced_weight = sum_red_cuda(gpu_vector_weight[mask, l, :])
                reduced_weighted_var = sum_red_cuda(gpu_weighted_var[mask, t,
                                                                     l, :])
                regmean[m][t][l] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var, gpu_vector_weight

    return regmean


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_weights_2D(mask, e1t, e2t):
    """Function that computes the regional weights for 2D vars in the cpu.

    Parameters
    ----------
    mask : float32
        Mask array containing a region mask.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.

    Returns
    -------
    weights: float32
        Masked array containing the weights for a given region.
    """
    weights = mask*e1t*e2t
    return weights


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weights_3D(mask, e1t, e2t, e3t):
    """Function that computes the regional weights for 3D vars in the cpu.

    Parameters
    ----------
    mask : float32
        Mask array containing a region mask.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e2t

    Returns
    -------
    weights: float32
        Masked array containing the weights for a given region.
    """
    weights = mask*e1t*e2t*e3t
    return weights


@cuda.jit
def variable_weight_cuda(var, e1t, e2t, masks, weighted_var, vector_weight):
    """Numba kernel executed on the gpu that computes the weights and
    weights a 2D variable for each region

    Parameters
    ----------
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    masks : float32
        Masked array containing region masks.
    weighted_var: float32
        Empty array to store the results.
    vector_var: float32
        Empty array to store the results.

    Returns
    -------
    weighted_var: float32
        Array containing the weighted variable for a given region.
    vector_weight: float32
        Array containing the weights for a given region.
    """
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    for bas in range(basins):
        vector_weight[bas, x] = 0.0
        weighted_var[bas, t, x] = 0.0
        vector_weight[bas, x] = e1t[j, i]*e2t[j, i]*masks[bas, j, i]
        weighted_var[bas, t, x] = var[t, j, i] * vector_weight[bas, x]


@cuda.jit
def variable_weight_cuda_3D(var, e1t, e2t, e3t, masks, weighted_var,
                            vector_weight):
    """Numba kernel executed on the gpu that computes the weights and
    weights a 3D variable for each region.

    Parameters
    ----------
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.
    masks : float32
        Masked array containing region masks.
    weighted_var: float32
        Empty array to store the results.
    vector_weight: float32
        Empty array to store the results.

    Returns
    -------
    weighted_var: float32
        Array containing the weighted variable for a given region.
    vector_weight: float32
        Array containing the weights for a given region.
    """
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in range(levels):
            x = i+var.shape[3]*(j+var.shape[2]*lev)
            vector_weight[bas, x] = 0.0
            vector_weight[bas, x] = (e1t[j, i] * e2t[j, i] *
                                     e3t[lev, j, i] * masks[bas, j, i])
            weighted_var[bas, t, x] = 0.0
            weighted_var[bas, t, x] = var[t, lev, j, i] * vector_weight[bas, x]


@cuda.jit
def variable_weight_cuda_level_3D(var, e1t, e2t, e3t, masks, weighted_var,
                                  vector_weight):
    """Numba kernel executed on the gpu that computes the weights and
    weights a 3D variable for each region.

    Parameters
    ----------
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    e3t: float32
        Masked array containing variable e3t.
    masks : float32
        Masked array containing region masks.
    weighted_var: float32
        Empty array to store the results.
    vector_weight: float32
        Empty array to store the results.

    Returns
    -------
    weighted_var: float32
        Array containing the weighted variable for a given region.
    vector_weight: float32
        Array containing the weights for a given region.
    """
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    x = i+j*var.shape[3]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in levels:
            vector_weight[bas, lev, x] = 0.0
            vector_weight[bas, lev, x] = (e1t[j, i] * e2t[j, i] *
                                          e3t[lev, j, i] * masks[bas, j, i])
            weighted_var[bas, t, lev, x] = (var[t, lev, j, i] *
                                            vector_weight[bas, lev, x])


@cuda.reduce
def sum_red_cuda(x, y):
    """Numba kernel executed on the gpu that performs a sum reduction:

    Parameters
    ----------
    x : float32
        Numpy array.
    y : float32
        Numpy array.

    Returns
    -------
    x+y : float64
        Reduction operation.
    """
    return x+y
